"""Modules for generating QA testbed jobs"""
# -*- coding: utf-8 -*-
# @author Ratnadip Adhikari

# exec(open("./arya-helpers/Codes/src/aryahelpers/aryatestbed/testbed.py").read())
from __future__ import absolute_import
import sys, os, json, logging
import warnings
warnings.filterwarnings('ignore')
os.system('clear') # print("\033c", end='')
# +++++++++++++++++
BASE_PATH = os.path.abspath("./arya-helpers/Codes/src")
extendPaths = [p for p in (os.path.dirname(BASE_PATH), BASE_PATH) if p not in sys.path]
sys.path = extendPaths + sys.path
# +++++++++++++++++
from joblib import Parallel, delayed
from datetime import datetime
from aryautils.storageutils import MySQLManager
from aryahelpers.config import CONFIG, CUSTOM_SAVE_PATH, ALL_FEATURES, CANDIDATE_TABLES
from aryahelpers.utils.textclean import PreprocessText
from aryahelpers.utils.genericutils import elapsedTime, save_files, merge_dicts, map_processing
from aryahelpers.utils.mysqlutils import MySQLHelpers
LOGGER = logging.getLogger(__name__)


class CreateQATestbed():
    """QA testbed creation class"""
    def __init__(self, job_ids=[]):
        self.testbed_jobs = []
        self.job_ids = job_ids
        self._ALL_FEATURES = set(k.lower() for k in ALL_FEATURES)
        self.toprocess_db_key, self.toprocess_db_name, self.toprocess_base_table, self.toprocess_jobid_key, \
            self.toprocess_guid_key = tuple(CANDIDATE_TABLES["toprocess"]["base"].values())
        self.expl_db_key, self.expl_db_name, self.expl_base_table, self.expl_jobid_key, \
            self.expl_guid_key, self.expl_key = tuple(CANDIDATE_TABLES["explanation"]["qa-reservoir"].values())
        self.toprocess_copy_table = CANDIDATE_TABLES["toprocess"]["copy"]["table"]
        self.expl_copy_table = CANDIDATE_TABLES["explanation"]["copy"]["table"]
        self.concise_info_keys = ['JobId', 'ClientCompany', 'JobTitle', 'SourceCount', 'PositiveCount', 'JTRCount']
        self.base_api = "http://192.168.38.99:1502/api/new?expl_type=facts"
    
    @staticmethod
    def _get_jtr_candidates(source_guids: set, jobid: int, db_key: str):
        """Get JTR candidates"""
        try:
            jtr_candidates = []
            base_query = "SELECT job_id, candidate_name, portal_candidate_id, recommendation_status_id, \
                CONVERT_TZ(jtr.updated_time,'+00:00','+05:30') AS updated_time FROM \
                    `arya`.`job_top_recommendations` jtr JOIN `arya`.`job_recommendations_metadata` jrm ON \
                        jtr.recommendation_id = jrm.recommendation_id WHERE job_id = "
            query_str = base_query + "{} ORDER BY candidate_score DESC;".format(jobid)
            query_res = MySQLManager.execute_query(query_str, (), **CONFIG[db_key]['arya'])
            for guid in source_guids:
                jtr_info = [d for d in query_res if d['portal_candidate_id'] == guid]
                jtr_info = sorted(jtr_info, key=lambda d: d['updated_time'], reverse=True)
                if jtr_info:
                    jtr_candidates.append(jtr_info[0])
        except Exception:
            jtr_candidates = []
        return jtr_candidates
        
    def _filter_sourced_jobs(self, refresh_data=False):
        """Checks each job id for existence in `toprocess_candidates` table and returns the existing ones.
        Also, copies all candidates to the `copy_table` based on the param: `refresh_data`."""
        sourced_jobs, copy_info = {}, []
        db_key, db_name, jobid_key, guid_key = self.toprocess_db_key, self.toprocess_db_name, \
            self.toprocess_jobid_key, self.toprocess_guid_key
        base_table, copy_table = self.toprocess_base_table, self.toprocess_copy_table
        # Select already existing job_ids from 'copy_table'
        try:
            existing_jobs = MySQLHelpers.get_contd_details(db_key, db_name, copy_table, jobid_key, guid_key)
        except Exception:
            existing_jobs = {}
        if not self.job_ids:
            [sourced_jobs.update({jid: merge_dicts((existing_jobs[jid], {
                "already_exists": True}))}) for jid in existing_jobs]
            return sourced_jobs
        # Now check each job for its existence in the 'toprocess_candidates' table
        for idx, jid in enumerate(self.job_ids):
            if jid in existing_jobs:
                sourced_jobs.update({jid: merge_dicts((existing_jobs[jid], {"already_exists": True}))})
            else:
                existing_basejob = MySQLHelpers.get_contd_details(db_key, db_name, base_table, jobid_key, guid_key, jid)
                if existing_basejob:
                    sourced_jobs.update({jid: merge_dicts((existing_basejob[jid], {"already_exists": False}))})
            if refresh_data:
                copy_info.append({"job_id": str(jid)})
            elif not sourced_jobs.get(jid, {}).get("already_exists"):
                copy_info.append({"job_id": str(jid)})
        # Copy the requisite jobs to 'copy_table'
        if not copy_info:
            copy_info = False
        if refresh_data:
            MySQLHelpers.transact_cands_from_tbl(base_table, db_key, db_name, "delete", copy_table, None, False, copy_info)
        MySQLHelpers.transact_cands_from_tbl(base_table, db_key, db_name, "copy", copy_table, None, False, copy_info)
        return sourced_jobs

    def _get_available_factkeys(self, expl_dict: dict):
        """Get the available (non-null) facts keys for a given eplanation dict"""
        all_factkeys = []
        [all_factkeys.append(k) for k, v in expl_dict.items() if k.lower() != 'summary'
         and ({k, k.lower()} & self._ALL_FEATURES) and v.get("Facts", [])]
        return all_factkeys
    
    def _extract_explcopied_jobs(self, refresh_data=False):
        """Extract set of jobs for which analyzed info are already copied"""
        db_key, db_name, jobid_key = self.expl_db_key, self.expl_db_name, self.expl_jobid_key
        if refresh_data:
            return set()
        try:
            existing_jobids = MySQLManager.execute_query("SELECT {} FROM {};".format(
                jobid_key, self.expl_copy_table), (), **CONFIG[db_key][db_name])
            return set([j for _dict in existing_jobids for k, j in _dict.items()])
        except Exception:
            return set()
    
    def _extract_explanation_facts(self, source_guids: set, job_id: int):
        """Extract existing explanation facts for the given job_id"""
        analyzed_contds, existing_facts, passed_count = [], [], 0
        db_key, db_name, expl_table, jobid_key, guid_key, expl_key = self.expl_db_key, self.expl_db_name, \
            self.expl_base_table, self.expl_jobid_key, self.expl_guid_key, self.expl_key
        try:            
            query_str = "SELECT *, CONVERT_TZ(UpdatedDateTime,'+00:00','+05:30') \
                AS UpdatedDateTime FROM {0} WHERE {1}={2} AND CandidateScore > 0;".format(
                    expl_table, jobid_key, job_id)                
            query_res = MySQLManager.execute_query(query_str, (), **CONFIG[db_key][db_name])
            # Filter analyzed contds w.r.t. source_guids
            for guid in source_guids:
                analyzed_info = [d for d in query_res if d[guid_key] == guid]
                analyzed_info = sorted(analyzed_info, key=lambda d: d['UpdatedDateTime'], reverse=True)
                if analyzed_info:
                    analyzed_contds.append(analyzed_info[0])
            passed_count = len(analyzed_contds)
            for idx, expl in enumerate(analyzed_contds):
                expl_keys = self._get_available_factkeys(json.loads(expl[expl_key]))
                _expl_keys = set(expl_keys)
                if set(existing_facts) <= _expl_keys:
                    existing_facts = expl_keys
                if (_expl_keys == self._ALL_FEATURES) or (_expl_keys == ALL_FEATURES):
                    break
        except Exception as error:
            LOGGER.error(error)
            LOGGER.error("Error in extracting existing explanation facts !!")
        return analyzed_contds, existing_facts, passed_count
    
    def _multiprocess_details(self, parallelize: bool, db_key: str, sourced_jobmap: dict, job_ids: list):
        """Obtain required DB-intensive details through multiprocessing"""
        args_list_facts = [(sourced_jobmap[jid]["contd_guids"], jid) for jid in job_ids]
        args_list_jtr = [(sourced_jobmap[jid]["contd_guids"], jid, db_key) for jid in job_ids]
        args_list_jobdetails = [(db_key, jid) for jid in job_ids]
        # Extract details through multiprocessing
        if parallelize:
            res_facts = Parallel(n_jobs=-1)(delayed(self._extract_explanation_facts)(*args) for args in args_list_facts)
            res_jtr = Parallel(n_jobs=-1)(delayed(self._get_jtr_candidates)(*args) for args in args_list_jtr)
            res_jobdetails = Parallel(n_jobs=-1)(delayed(MySQLHelpers.extract_arya_jobdetails)(*args) for args in args_list_jobdetails)
        else:
            res_facts = map_processing(self._extract_explanation_facts, args_list_facts, True)
            res_jtr = map_processing(self._get_jtr_candidates, args_list_jtr, True)
            res_jobdetails = map_processing(MySQLHelpers.extract_arya_jobdetails, args_list_jobdetails, True)
        # Create dict of the extracted details
        extracted_facts = {jid: res_facts[idx] for idx, jid in enumerate(job_ids)}
        extracted_jtr = {jid: len(res_jtr[idx]) for idx, jid in enumerate(job_ids)}
        extracted_jobdetails = {jid: res_jobdetails[idx] for idx, jid in enumerate(job_ids)}
        return extracted_facts, extracted_jtr, extracted_jobdetails
        
    def _summarize_testbed_jobs(self, concise_info=False):
        """Get a summarized (condensed) form of testbed job details"""
        if self.testbed_jobs and concise_info:
            self.testbed_jobs = sorted([tuple({k: v for k, v in d.items() if k in self.concise_info_keys}.values())
                                        for d in self.testbed_jobs], key=lambda x: x[0], reverse=True)
            self.testbed_jobs = {str(k[0]): k[1:] for k in self.testbed_jobs}
        
    @elapsedTime
    def cleanup_custom_tables(self, keep=False):
        """
        Delete undesirable jobs data from the custom table. The param `keep=False` refers deletion of
        the given jobs, whereas `keep=True` refers to keep the given jobs and delete all others.
        """
        not_str = 'NOT' if keep else ''
        db_key, db_name, job_ids = self.toprocess_db_key, self.toprocess_db_name, tuple(set(self.job_ids))
        cleanup_status = {"toprocess": False, "analyzed_info": False}
        for table in (self.toprocess_copy_table, self.expl_copy_table):
            jobid_key = self.toprocess_jobid_key if table == self.toprocess_copy_table else self.expl_jobid_key
            status_key = 'toprocess' if table == self.toprocess_copy_table else 'analyzed_info'
            query_str = "DELETE FROM {0} WHERE {1} {2} IN {3};".format(table, jobid_key, not_str, job_ids)
            query_str = PreprocessText().spaces_clean(query_str)
            cleanup_status[status_key] = MySQLHelpers.mysql_query_execute(query_str, (), **CONFIG[db_key][db_name])
        return cleanup_status
    
    @elapsedTime
    def extract_testbed_candidates(self, to_extract='both', **kwargs):
        """Extract all testbed candidates from custom toprocess & analyzed info tables as pickle zip files.
        
        Args:
            to_extract (str, optional): Type of candidates to extract: 'both' (default), 'toprocess' or 'analyzed'.
            save_path (bool, optional): The path to save extracted candidates. Defaults to ``False``.
            **kwargs (optional): Any other optional args for `save_files` and `MySQLHelpers.transact_cands_from_tbl`.
        
        Returns:
            List of extracted candidates.
        """
        exclude_kwargs = ['from_table', 'db_key', 'db_name', 'transact_type', 'to_table',
                          'copy_info', 'file_name', 'save_file_type']
        kwargs = {k: v for k, v in kwargs.items() if k not in exclude_kwargs}
        extracted_toprocess, extracted_analyzed = [], []
        if to_extract in ('toprocess', 'both'):
            extracted_toprocess, _ = MySQLHelpers.transact_cands_from_tbl(
                self.toprocess_copy_table, self.toprocess_db_key, self.toprocess_db_name, 'extract',
                copy_info=[], file_name='custom_toprocess', save_file_type='p_zip', **kwargs)
        if to_extract in ('analyzed', 'both'):
            extracted_analyzed, _ = MySQLHelpers.transact_cands_from_tbl(
                self.expl_copy_table, self.expl_db_key, self.expl_db_name, 'extract',
                copy_info=[], file_name='custom_analyzed_info', save_file_type='p_zip', **kwargs)
        return extracted_toprocess, extracted_analyzed
    
    @elapsedTime
    def create_testbed(self, refresh_data=False, concise_info=False, parallelize=True, save_path=False):
        """Main module for creating testbed jobs"""
        refresh_data = False if not self.job_ids else refresh_data
        sourced_jobs = self._filter_sourced_jobs(refresh_data)  # collect jobs which exist in toprocess table
        existing_jobids = self._extract_explcopied_jobs(refresh_data)
        job_ids = set(sourced_jobs) if self.job_ids else set(existing_jobids)
        sourced_jobs = {k: v for k, v in sourced_jobs.items() if k in job_ids}
        job_ids = sorted(list(job_ids), reverse=True)
        db_key, db_name, jobid_key, base_table, copy_table = self.expl_db_key, self.expl_db_name,\
            self.expl_jobid_key, self.expl_base_table, self.expl_copy_table
        facts_details, jtr_counts, arya_jobdetails = self._multiprocess_details(parallelize, db_key, sourced_jobs, job_ids)
        copy_info, analyzed_contds, jobs_record = [], [], []
        try:
            for jid in job_ids:
                contd_details, existing_facts, passed_count = facts_details[jid]
                if not existing_facts:
                    continue
                if jid not in existing_jobids:
                    copy_info.append({jobid_key: str(jid)})
                ccompany_job_title = [arya_jobdetails[jid].get(k, '') for k in ('ClientCompany', 'JobTitle')]
                if any(ccompany_job_title) and (ccompany_job_title in jobs_record):
                    continue  # Record only unique jobs w.r.t. -- ClientCompany-JobTitle
                expl_jobdetails = {
                    "SourceCount": sourced_jobs[jid]['source_count'],
                    "PositiveCount": passed_count,
                    "JTRCount": jtr_counts[jid],
                    "ExplanationKeys": existing_facts,
                    "AvailableExplanations": len(existing_facts),
                    "ExplanationAPI": self.base_api + "&job_id={}&num_candidates=20".format(jid)
                }
                if jid not in existing_jobids:
                    analyzed_contds.extend(contd_details)
                self.testbed_jobs.append(merge_dicts(({"JobId": jid}, arya_jobdetails[jid], expl_jobdetails)))
                jobs_record.append(ccompany_job_title)
            # +++++++++++++++++
            # Copy all candidates details for whom there are explanation facts
            if not copy_info:
                copy_info = False
            MySQLHelpers.transact_cands_from_tbl(base_table, db_key, db_name, "delete", copy_table, copy_info=copy_info)
            if analyzed_contds:
                MySQLHelpers.insert_bulk_data(analyzed_contds, db_key, db_name, base_table, copy_table)
            self._summarize_testbed_jobs(concise_info)
            save_files(self.testbed_jobs, save_path, 'explanation_jobs_details', '.json', indent=4, default=str)
        except Exception as error:
            LOGGER.error(error)
            LOGGER.error("Unexpected Error in creating QA testbed !!")
            raise error
       

if __name__ == '__main__':
    print('python version:', sys.version)   
    print('cwd:', os.getcwd())
    # _testbed = CreateQATestbed(job_ids=[377534, 377522, 377407, 377137])
    # _testbed.create_testbed(refresh_data=False, parallelize=True, concise_info=True, save_path=CUSTOM_SAVE_PATH)
    # print(json.dumps(_testbed.testbed_jobs, indent=4, default=str))



import json
import pickle
from pkgutil import get_data

config_folder, data_folder = 'aryahelpers.configuration', 'aryahelpers.data'
CONFIG = json.loads(get_data(config_folder, 'config.helpers.json').decode())
HEADERS = {'Content-Type': 'application/json'}
CUSTOM_SAVE_PATH = ''.join(CONFIG['CUSTOM_SAVE_PATH'])
ENGLISH_WORDS = set(pickle.loads(get_data(data_folder, 'english_words.p'))["ENGLISH_WORDS"])
CANDIDATE_TABLES = CONFIG["candidate_tables"]
COPY_JOB_LIMIT = 20
FACT_TOKENS = {'ExplanationFacts', 'Explanation', 'facts', 'Facts'}
ALL_FEATURES = {'Summary', 'Skill', 'Role', 'Occupation', 'Company', 'Industry', 'Education', 'Experience'}

# Bucket dict
BUCKET_DICT = {
    1: ["0-50", "Small"],
    2: ["50-250", "Small"],
    3: ["250-500", "Small"],
    4: ["500-1K", "Medium"],
    5: ["1K-5K", "Medium"],
    6: ["5K-10K", "Medium"],
    7: ["10K-50K", "Large"],
    8: ["50K-100K", "Large"],
    9: ["100K-500K", "Large"],
    10: ["500K-1M", "Large"],
    11: ["1M-10M", "Very large"],
    12: ["10M-100M", "Very large"],
    13: ["100M-500M", "Very large"],
    14: ["500M-1B", "Very large"],
    15: ["1B+", "Very large"]
}

# Companies specific service stopwords
company_stopwords = [
    'agreement', 'amendment', 'bill', 'biz', 'blank', 'bv', 'client', 'clubs', 'co', 'com', 'comp',
    'companies', 'company', 'confidential', 'contract', 'corp', 'corporate', 'corporation', 'dba', 'government',
    'govt', 'group', 'http', 'https', 'inc', 'including', 'incorp', 'incorporat', 'incorporate', 'incorporated',
    'incorporation', 'industries', 'international', 'intl', 'intnl', 'license', 'limited', 'llc', 'llp', 'ltd',
    'machines', 'maintenance', 'membership', 'name', 'ny', 'office', 'oh', 'private', 'pte', 'pvt', 'rent', 's',
    'san', 'service', 'services', 'st', 'unknown', 'utilities', 'utility', 'web', 'website', 'working', 'www',
    'yrs', 'yr', 'mos', 'mo', 'fulltime', 'parttime', 'permanent', 'temporary'
]


"""arya-helpers testing script"""
# -*- coding: utf-8 -*-
# @author Ratnadip Adhikari

# execute the script
# exec(open("./arya-helpers/Codes/src/helpers_testing.py").read())

import sys, os, json
import pandas as pd
from functools import reduce
from importlib import reload, import_module
from IPython import get_ipython
# get_ipython().magic('reset -f')
import warnings
warnings.filterwarnings('ignore') ## Ignore warnings
os.system('clear') # print("\033c", end='')  ## clear screen
# globals().clear(); locals().clear() ## for resetting python interpreter memory
# +++++++++++++++++
# Boilerplate code for resolving relative imports
BASE_PATH = os.path.abspath("./arya-helpers/Codes/src")
extendPaths = [p for p in (os.path.dirname(BASE_PATH), BASE_PATH) if p not in sys.path]
sys.path = extendPaths + sys.path
# +++++++++++++++++
from aryahelpers.config import *
from aryahelpers.aryatestbed import testbed
from aryahelpers.utils import genericutils, mysqlutils, textclean
from aryautils.storageutils import MySQLManager
reload(testbed)
reload(sys.modules['aryahelpers.utils.genericutils'])
reload(sys.modules['aryahelpers.utils.mysqlutils'])
reload(sys.modules['aryahelpers.utils.textclean'])
from aryahelpers.utils.genericutils import *
from aryahelpers.utils.mysqlutils import MySQLHelpers
from aryahelpers.utils.textclean import PreprocessText
from aryahelpers.aryatestbed.testbed import CreateQATestbed
from aryahelpers.joboperations.aryajobapi import AryaJobOperations
# ==================================

# ==================================
if False:
    # =================
    # QA job MySql operations
    MySQLHelpers.qa_job_mysql_operations(job_id=373387, query_seq=3, copy_info=False, to_print=True)
    # -*--*--*--*--*-
    # MySql table queries
    query_strs = [
        "SELECT COUNT(*) FROM custom_toprocess_candidates;",
        "SELECT COUNT(DISTINCT job_id) FROM custom_toprocess_candidates;",
        "SELECT COUNT(*) FROM custom_toprocess_analyzed_info;",
        "SELECT COUNT(DISTINCT jobId) FROM custom_toprocess_analyzed_info;"
    ]
    MySQLManager.execute_query(query_strs[0], (), **CONFIG['database']['candidate_reservoir'])
    # -*--*--*--*--*-
    query_strs = [
        "TRUNCATE TABLE custom_toprocess_candidates;",
        "DROP TABLE custom_toprocess_candidates;",
        "TRUNCATE TABLE custom_toprocess_analyzed_info;",
        "DROP TABLE custom_toprocess_analyzed_info;"
    ]
    MySQLHelpers.mysql_query_execute(query_strs[0], (), **CONFIG['database']['candidate_reservoir'])
    # -*--*--*--*--*-
    query_str = "SELECT DISTINCT(job_id) FROM custom_toprocess_candidates;"
    job_ids = MySQLManager.execute_query(query_str, (), **CONFIG['database']['candidate_reservoir'])
    job_ids = sorted(list(set([d['job_id'] for d in job_ids])), reverse=True)
    
    # =================
    # Determine sourced jobs
    filePath = "./arya-helpers/Codes/src/aryahelpers/data"
    testbed_jobs = import_file(os.path.join(filePath, "QA_testbed_jobs.json"))
    updated_jobs = {}
    for jid, source_details in testbed_jobs.items():
        query_str = "SELECT COUNT(*) FROM toprocess_candidates WHERE job_id={};".format(jid)
        query_res = MySQLManager.execute_query(query_str, (), **CONFIG['database']['candidate_reservoir'])
        source_count = query_res[0]["COUNT(*)"]
        jtr_candidates = MySQLHelpers.get_TopN_JTR(jid, 'database')
        if source_count and len(jtr_candidates) and len(source_details) == 3:
            updated_jobs.update({jid: source_details[:1] + [source_count, "Sourced"]})
        else:
            updated_jobs.update({jid: source_details[:1] + [source_count]})
    # -*--*--*--*--*-
    # Save updated sourced job details
    save_files(updated_jobs, filePath, 'QA_testbed_jobs', '.json', indent=4)
    
    # =================
    # Call `MySQLHelpers.transact_cands_from_tbl`
    params = {
        "from_table": 'custom_toprocess_analyzed_info',
        "db_key": 'database',
        "db_name": 'candidate_reservoir',
        "transact_type": 'extract',
        "to_table": None,
        "select_cols": None,
        "to_truncate": False,
        "copy_info": [],
        "where_clause": 'CandidateScore >= 2',
        "save_path": False,
        "file_name": 'extracted_info',
        "save_file_type": 'p_zip',
        "file_name_suffix": '',
        "indent": 4,
        "default": str
    }
    extractCands, _ = MySQLHelpers.transact_cands_from_tbl(**params)
    # -*--*--*--*--*-
    # Load saved table data
    data_path = os.path.join(CUSTOM_SAVE_PATH, 'extracted_testbed', 'custom_analyzed_info.zip')
    contds_info = import_file(data_path)

    # =================
    # Create Arya sourced jobs record
    filter_dict = {
        "JobCreatedEmail": [
            "sreevally.pasumarthy@leoforce.com",
            "dsqatesting@leoforce.com"
        ],
        "JobCreatedDate": [
            "01-Nov-2023, 00:00:00 IST",
            "07-Dec-2023, 00:00:00 IST"
        ]
    }
    joboperations_obj = AryaJobOperations(job_ids=range(387000, 387830))
    joboperations_obj.arya_sourced_jobs_record(filter_dict, CUSTOM_SAVE_PATH)
    
    # =================
    # Clients mapping operations for users
    save_path = "/home/ratnadipadhikari/OneDrive/Leoforce/Candidate_Explanation_V2/Experiment_Stuffs/Copy_Job"
    clients_mapping = MySQLHelpers().get_clientsmapping_for_user(
        'database', 'dsqatesting@leoforce.com', 'user_email', save_path)
    # -*--*--*--*--*-
    MySQLHelpers()._get_user_info_from_db('database', 'dsqatesting@leoforce.com', id_type='user_email')
    MySQLHelpers().fetch_or_add_client_for_user('database', 'zomato', 'dsqatesting@leoforce.com', 'user_email')
    MySQLHelpers().delete_clients_from_user('database', ['zomato', 'a2aa'], 'dsqatesting@leoforce.com', 'user_email')
# ==================================

if __name__ == "__main__":
    print('python version:', sys.version) 
    print('cwd:', os.getcwd())



"""Flask api to access candidates explanations from DB"""
# http://192.168.38.99:1502/api/qa?expl_type=facts&job_id=359229&name=Vinay%20Pratap
# http://192.168.38.99:1502/api/qa?expl_type=facts&job_id=371151&recommendation_status_id=3
# http://192.168.38.99:1502/api/qa?expl_type=company&job_id=359228&num_candidates=10
# http://192.168.38.99:1502/api/staging?job_id=199436&num_candidates=20


import sys, os, json
BASE_PATH = os.path.abspath("./arya-helpers/Codes/src")
extendPaths = [p for p in (os.path.dirname(BASE_PATH), BASE_PATH) if p not in sys.path]
sys.path = extendPaths + sys.path
from copy import deepcopy
from aryautils.storageutils import MySQLManager
from aryahelpers.config import CONFIG, ALL_FEATURES, FACT_TOKENS, CANDIDATE_TABLES
from aryahelpers.utils.genericutils import merge_dicts, flatten_iterable
from flask import Flask, request, jsonify
import logging
LOGGER = logging.getLogger(__name__)

app = Flask(__name__)


class AccessExplanations(object):
    """Access candidates explanations from DB"""
    def __init__(self, **kwargs):
        self._ALL_FEATURES = set(k.lower() for k in ALL_FEATURES)
        [setattr(self, k, v) for k, v in kwargs.items()]
        # job_id, num_candidates, candidate_name, candidate_guid
    
    @staticmethod
    def get_jtr_candidates(jobid, db_key):
        """Get JTR candidates"""
        base_query = "SELECT job_id, candidate_name, portal_candidate_id, recommendation_status_id \
            FROM `arya`.`job_top_recommendations` jtr JOIN `arya`.`job_recommendations_metadata` jrm \
                ON jtr.recommendation_id = jrm.recommendation_id WHERE job_id = "
        query_str = base_query + "{} ORDER BY candidate_score DESC;".format(jobid)
        jtr_candidates = MySQLManager.execute_query(query_str, (), **CONFIG[db_key]['arya'])
        return jtr_candidates
    
    def get_explanation_facts(self, candidate_details: dict):
        """Filter explanations, based on `explanation_type`"""
        explanation_type = self.explanation_type  # 'company', 'skill', 'role', 'occupation',
                                                  # 'industry', 'experience', 'facts'
        filtered_explanations = {}
        if explanation_type:
            if explanation_type != 'facts':
                filtered_explanations = candidate_details.get(
                    explanation_type, candidate_details.get(explanation_type.title(), {}))
            else:
                all_explkeys = list(candidate_details.keys())
                all_explkeys = {k: k if k.lower() != 'summary' else f'#{k}' for k in all_explkeys if isinstance(
                    candidate_details[k], dict) and (set(candidate_details[k].keys()) & FACT_TOKENS)}
                filtered_explanations["ExplanationFacts"] = {}
                for _explkey in all_explkeys:
                    for _fact_key in FACT_TOKENS:
                        get_facts = candidate_details.get(_explkey, {}).get(_fact_key, [])
                        if isinstance(get_facts, list):
                            get_facts = [_fact.get('Fact', '') + _fact.get('fact', '') if isinstance(
                                _fact, dict) else _fact for _fact in get_facts]
                        if get_facts:
                            filtered_explanations["ExplanationFacts"][all_explkeys[_explkey]] = get_facts
                            break
        else:
            return candidate_details
        return filtered_explanations
    
    def extract_analyzed_candidates(self, expl_tbl_key, job_id=None, expl_type=None):
        """Extract analyzed candidates details from the respective db"""
        cand_details = []
        db_key, db_name, table_name, jobid_key, contd_guid, expl_key = tuple(
            CANDIDATE_TABLES["explanation"][expl_tbl_key].values())
        if expl_tbl_key == 'staging':
            query_str = "SELECT {}, {}, {} FROM {} WHERE {}={}".format(
                jobid_key, contd_guid, expl_key, table_name, jobid_key, job_id)
        else:
            # Add the respective Score
            score_keys = ["CandidateScore"]
            if expl_type in self._ALL_FEATURES:
                score_keys += [] if expl_type in ['facts', 'summary'] else [
                    "{}Score".format(expl_type.title())] if expl_type != 'role' else ['RoleLevelScore']
            else:
                score_keys += ['CompanyScore', 'RoleLevelScore', 'OccupationScore', 'EducationScore',
                               'ExperienceScore', 'IndustryScore', 'SkillScore']
            query_str = "SELECT {}, {}, {}, CONVERT_TZ(UpdatedDateTime,'+00:00','+05:30') \
                AS UpdatedDateTime, {} FROM {} WHERE {}={}".format(
                    jobid_key, ", ".join(score_keys), contd_guid, expl_key, table_name, jobid_key, job_id)
        cand_details = MySQLManager.execute_query(query_str, (), **CONFIG[db_key][db_name])
        cand_details = [{k: v for k, v in d.items() if v or (k not in score_keys)} for d in cand_details]
        for idx, d in enumerate(cand_details):
            cand_details[idx][expl_key] = deepcopy(json.loads(d[expl_key]))
            cand_details[idx][expl_key] = self.get_explanation_facts(cand_details[idx][expl_key])
        return cand_details
    
    def filter_candidates(self, contd_details: list, filter_keys: tuple):
        """Filter `contd_details` with respect to `filter_keys`"""
        filtered_details = contd_details.copy()
        num_candidates = self.num_candidates
        candidate_name = self.candidate_name
        candidate_guid = self.candidate_guid
        recommendation_status_id = self.recommendation_status_id  #[1, 2, 3]
        if num_candidates:
            filtered_details = [d for d in filtered_details[:num_candidates]]
        if candidate_name:
            filtered_details = [d for d in filtered_details if d[filter_keys[0]] == candidate_name]
        if candidate_guid:
            filtered_details = [d for d in filtered_details if d[filter_keys[1]] == candidate_guid]
        if recommendation_status_id:
            filtered_details = [d for d in filtered_details if d[filter_keys[2]] == recommendation_status_id]
        return filtered_details
    
    def access_explanations(self):
        """Access explanations, based on specific filtering conditions"""
        job_id, base_env = self.job_id, self.base_env
        explanation_type = self.explanation_type

        # Access candidates explanations and other details
        cand_explanations, filter_guids = [], []
        expl_tbl_key = 'qa-reservoir' if base_env == 'qa' else 'copy' if base_env == 'new' else 'staging'
        cand_details = self.extract_analyzed_candidates(expl_tbl_key, job_id, explanation_type)
        if not cand_details and (base_env in ['qa', 'new']):
            expl_tbl_key = 'qa-arya'
            cand_details = self.extract_analyzed_candidates(expl_tbl_key, job_id, explanation_type)
        
        # Filter cand_guids
        db_key, guid_key = [CANDIDATE_TABLES["explanation"][expl_tbl_key][k] for k in ('db_key', 'guid_key')]
        jtr_candidates = self.get_jtr_candidates(job_id, db_key)
        if jtr_candidates:
            filter_guids = [(d['candidate_name'], d['portal_candidate_id'],
                             d['recommendation_status_id']) for d in jtr_candidates]
            filter_guids = self.filter_candidates(filter_guids, (0, 1, 2))
            # Access explanations
            for cand_info in filter_guids:
                subset_cands = [d for d in cand_details if d[guid_key] == cand_info[1]]
                if guid_key == 'portalGuid':
                    subset_cands = sorted(subset_cands, key=lambda d: d['UpdatedDateTime'], reverse=True)
                if subset_cands:
                    subset_cands = merge_dicts(({
                        "CandidateName": cand_info[0], "recommendation_status_id": cand_info[2]}, subset_cands[0]))
                    cand_explanations.append(subset_cands)
        else:
            # if jtr_candidates list is empty, e.g. manually entered info
            cand_details = [d for d in cand_details if d.get("CandidateScore", 0) >= 0]
            cand_details = sorted(cand_details, key=lambda x: x.get("CandidateScore", 0), reverse=True)
            cand_details = self.filter_candidates(cand_details, ('', guid_key, ''))
            cand_explanations.extend(cand_details)
        cand_explanations = flatten_iterable(cand_explanations)
        return cand_explanations
    

@app.route('/api/<base_env>', methods=['GET'], strict_slashes=False)
def main_func(base_env=None):
    try:
        response = []
        # Access input params
        job_id = request.args.get('job_id')
        num_candidates = request.args.get('num_candidates')
        if num_candidates:
            num_candidates = int(num_candidates)
        candidate_name = request.args.get('name')
        candidate_guid = request.args.get('guid')
        recommendation_status_id = request.args.get('recommendation_status_id')
        explanation_type = request.args.get('expl_type', 'facts').lower()
        if recommendation_status_id:
            recommendation_status_id = int(recommendation_status_id)
        accessor = AccessExplanations(
            job_id=job_id, num_candidates=num_candidates, candidate_name=candidate_name,
            candidate_guid=candidate_guid, recommendation_status_id=recommendation_status_id,
            explanation_type=explanation_type, base_env=base_env)
        response = accessor.access_explanations()
    except Exception as error:
        LOGGER.error('Failed in accessing candidates explanations from DB !')
        LOGGER.exception(error)
    return jsonify(response)
  

if __name__ == '__main__':
    app.run("0.0.0.0", port=1502)



"""Flask api to peroform automated Arya job operations"""

import sys
import os
import requests
import base64
BASE_PATH = os.path.abspath("./arya-helpers/Codes/src")
extendPaths = [p for p in (os.path.dirname(BASE_PATH), BASE_PATH) if p not in sys.path]
sys.path = extendPaths + sys.path
from datetime import datetime
from joblib import Parallel, delayed
from aryautils.storageutils import MySQLManager
from aryahelpers.config import CONFIG, COPY_JOB_LIMIT, CANDIDATE_TABLES
from aryahelpers.utils.genericutils import elapsedTime, merge_dicts, save_files
from aryahelpers.utils.mysqlutils import MySQLHelpers
from flask import Flask, request, abort, jsonify
import logging
LOGGER = logging.getLogger(__name__)

app = Flask(__name__)


class AryaJobOperations():
    """Arya job operations class"""
    def __init__(self, **kwargs):
        self.db_key, self.db_name, self.base_table, self.jobid_key, self.guid_key = \
            tuple(CANDIDATE_TABLES["toprocess"]["base"].values())
        self.copy_table = CANDIDATE_TABLES["toprocess"]["copy"]["table"]
        self.arya_job_url = f"{CONFIG['arya_job_urls']['jobs_v3.1']}"
        [setattr(self, k, v) for k, v in kwargs.items()]
        # job_ids, source_user_email, target_user_email, target_jobs_status, to_retain_job

    @staticmethod
    def validate_copy_jobs_payload(payload):
        """Validate the payload for copy jobs"""
        if payload and payload['job_ids'] and len(payload['job_ids']) <= COPY_JOB_LIMIT and \
            payload['source_user_email'] and payload['target_user_email']:
            return True
        return False
    
    @staticmethod
    def form_headers(user_email):
        """Form headers from `user_email`"""
        headers = {}
        bs4_email = base64.b64encode(user_email.encode('utf-8')).decode('utf-8')
        headers['Authorization'] = f"AryaKey {CONFIG['arya_api_auth']['application_id']};{CONFIG['arya_api_auth']['application_key']};{bs4_email}"
        headers['accept'] = "application/octet-stream"
        headers['Content-Type'] = "application/json"
        return headers

    def make_request(self, req_method, url, url_name=None, user_email=None, payload=None, headers=None):
        """Make GET/POST request"""
        if user_email:
            headers = AryaJobOperations.form_headers(user_email)
        request_resp = requests.get(url, json=payload, headers=headers) if req_method == 'GET' \
            else requests.post(url, json=payload, headers=headers)
        print(f'{url_name} response status code: {request_resp.status_code}')
        if not request_resp.content:
            return
        return request_resp.json()
    
    def copy_single_job(self, job_id):
        """Copy a single job from source to target user"""
        copyjob_info = {"source_job": job_id, "target_job": None}
        job_status_id = getattr(self, 'target_jobs_status', 2)
        to_retain_job = getattr(self, 'to_retain_job', False)
        source_job = self.make_request('GET', f"{self.arya_job_url}/{job_id}", 'GET JOB', self.source_user_email)
        if not source_job:
            return copyjob_info
        print(f'Fetched job: {job_id}')
        client_mapping = MySQLHelpers.fetch_or_add_client_for_user(
            self.db_key, source_job['Client'], self.target_user_email, 'user_email')
        if not client_mapping:
            return copyjob_info
        source_job['StatusId'], source_job['ClientId'], source_job['AssignedTo'] = \
            job_status_id, client_mapping['ClientID'], [self.USER['UserGuid']]
        target_job = self.make_request('POST', self.arya_job_url, 'CREATE JOB', self.target_user_email, source_job)
        if target_job:
            target_job_id = target_job['JobId']
            if to_retain_job:
                MySQLHelpers.transact_to_jobsretain_tbl(self.db_key, 'candidate_reservoir', [target_job_id], 'insert')
            print(f"Created job: {target_job_id}")
            copyjob_info = {"source_job": job_id, "target_job": target_job_id}
        return copyjob_info

    @elapsedTime    
    def copy_jobs_from_source_to_target(self, payload):
        """Module for copying jobs from source user to target user"""
        self.USER = MySQLHelpers._get_user_info_from_db(self.db_key, self.target_user_email, 'user_email')
        self.USER = self.USER[0] if self.USER else {}
        self.copyjobs_info = [self.copy_single_job(jid) for jid in self.job_ids]
        # self.copyjobs_info = Parallel(n_jobs=-1)(delayed(self.copy_single_job)(jid) for jid in self.job_ids)

    def _filter_jobs_record(self, jobs_record: dict, filter_dict: dict):
        """Filter jobs record based on the `filter_dict` param"""
        filtered_jobs_record = {}
        filter_dict = {k: [datetime.strptime(dt, "%d-%b-%Y, %H:%M:%S IST") for dt in v]
                       if 'date' in k.lower() else v for k, v in filter_dict.items()}
        for jid, jrecord in jobs_record.items():
            filterFlag = True  # filter flag
            for fkey, frange in filter_dict.items():
                if 'date' in fkey.lower():
                    if not (min(frange) <= datetime.strptime(jrecord[fkey], "%d-%b-%Y, %H:%M:%S IST") <= max(frange)):
                        filterFlag = False
                        break
                elif all([isinstance(v, int) for v in frange]):
                    if not (min(frange) <= jrecord[fkey] <= max(frange)):
                        filterFlag = False
                        break
                elif jrecord[fkey] not in frange:
                    filterFlag = False
                    break
            if filterFlag:
                filtered_jobs_record.update({jid: jobs_record[jid]})
        return filtered_jobs_record
    
    @elapsedTime
    def arya_sourced_jobs_record(self, filter_dict={}, save_path=False):
        """Derive few stats regarding the Arya sourced jobs, e.g. -- 
        - does the job exist in the retained table
        - does it exist in the toprocess and/or custom toprocess tables
        - job creation & last update dates
        """
        arya_jobs_record, job_ids = {}, sorted(list(self.job_ids), reverse=True)
        self.sourced_jobs_record = {}
        try:
            db_key, db_name, jobid_key = self.db_key, self.db_name, self.jobid_key
            base_table, copy_table = self.base_table, self.copy_table
            toprocess_jobs = MySQLHelpers.get_all_distinct_jobs(db_key, db_name, base_table, jobid_key)
            custom_toprocess_jobs = MySQLHelpers.get_all_distinct_jobs(db_key, db_name, copy_table, jobid_key)
            retained_jobs = MySQLHelpers.transact_to_jobsretain_tbl(db_key, db_name, job_ids, 'check')
            # Get Arya job details & JTR counts
            arya_jobdetails = Parallel(n_jobs=-1)(delayed(MySQLHelpers.extract_arya_jobdetails)(db_key, jid) for jid in job_ids)
            jtr_contds = Parallel(n_jobs=-1)(delayed(MySQLHelpers.get_TopN_JTR)(db_key, jid) for jid in job_ids)
                
            # Form the sourced jobs record
            unique_jobs_record = []
            for idx, jid in enumerate(job_ids):
                if not arya_jobdetails[idx]:
                    continue
                ccompany_job_title = [arya_jobdetails[idx].get(k, '') for k in ('ClientCompany', 'JobTitle')]
                if any(ccompany_job_title) and (ccompany_job_title in unique_jobs_record):
                    continue  # Record only unique jobs w.r.t. -- ClientCompany-JobTitle
                existence_records = {
                    "is_in_toprocess": jid in toprocess_jobs,
                    "is_in_custom_toprocess": jid in custom_toprocess_jobs,
                    "is_ratined_job": jid in retained_jobs,
                    "JTRCount": len(jtr_contds[idx]),
                }
                job_record = merge_dicts((existence_records, arya_jobdetails[idx]))
                arya_jobs_record.update({jid: job_record})
                unique_jobs_record.append(ccompany_job_title)
            self.sourced_jobs_record = self._filter_jobs_record(arya_jobs_record, filter_dict)  # filter jobs record
            save_files(self.sourced_jobs_record, save_path, 'sourced_jobs_record', '.json', indent=4, default=str)
        except Exception as error:
            LOGGER.error(error)
            LOGGER.error("Unexpected Error in gathering Arya sourced jobs record !!")
            raise error


@app.route('/copy_jobs', methods=['POST'])
def copy_jobs():
    """The copy job route. Example request payload is shown below.
    {
        "job_ids": [386227, 383180, 382849],
        "source_user_email": "sreevally.pasumarthy@leoforce.com",
        "target_user_email": "dsqatesting@leoforce.com",
        "target_jobs_status": 1,
        "to_retain_job": false
    }
    """
    try:
        response = []
        payload = request.json
        joboperations_obj = AryaJobOperations(**payload)
        if not joboperations_obj.validate_copy_jobs_payload(payload):
            abort(400, 'Invalid request, job ids, source and target emails are required to process the request.')
        joboperations_obj.copy_jobs_from_source_to_target(payload)
        response = joboperations_obj.copyjobs_info
    except Exception as error:
        LOGGER.error('Failed in copying jobs from source to target user !')
        LOGGER.exception(error)
    return jsonify(response)

@app.route('/jobs_record', methods=['POST'])
def jobs_record():
    """The Arya sourced jobs route. Example request payload is shown below.
    {
        "job_ids": [387814, 387815, 387816, 387817],
        "filter_dict": {
            "JobCreatedEmail": [
                "sreevally.pasumarthy@leoforce.com",
                "dsqatesting@leoforce.com"
            ],
            "JobCreatedDate": [
                "01-Nov-2023, 00:00:00 IST",
                "07-Dec-2023, 00:00:00 IST"
            ]
        },
        "save_path": "/path_to_save"
    }
    """
    try:
        response = {}
        payload = request.json
        job_ids = payload.get("job_ids", [])
        filter_dict = payload.get("filter_dict", {})
        save_path = payload.get("save_path", False)
        joboperations_obj = AryaJobOperations(job_ids=job_ids)
        joboperations_obj.arya_sourced_jobs_record(filter_dict, save_path)
        response = joboperations_obj.sourced_jobs_record
    except Exception as error:
        LOGGER.error('Failed in copying jobs from source to target user !')
        LOGGER.exception(error)
    return jsonify(response)


if __name__ == '__main__':
    app.run("0.0.0.0", port=1511)



import logging
import logging.handlers
from aryaesloghandler.handlers import AryaElasticLogHandler


def configure_logging(apps, config=None):
    if not apps:
        apps = [""]
    logging.basicConfig(format='%(asctime)s %(filename)s %(levelname)s: %(message)s',
                        datefmt='%m/%d/%Y %I:%M:%S %p',
                        level=logging.INFO)
    es_handler = AryaElasticLogHandler(
        hosts=[{
            'host': config['logs_es_host'],
            'port': config['logs_es_port']
        }],
        use_ssl=False if config['logs_es_use_ssl'] == 'False' else True,
        # Can configure corresponding authentication authority
        auth_type=AryaElasticLogHandler.AuthType.NO_AUTH,
        es_index_name=config['logs_es_index'],
        # One index per month
        index_name_frequency=AryaElasticLogHandler.IndexNameFrequency.MONTHLY,
        flush_frequency_in_sec=15,
        # Additional environmental identification
        raise_on_indexing_exceptions=False)
    es_handler.setLevel(level=logging.INFO)
    for app_name in apps:
        logger = logging.getLogger(app_name)
        logger.addHandler(es_handler)



"""Setup for aryahelpers modules"""
from setuptools import setup, find_packages

BASE_REPO = 'https://gitlab.com/aryacoreprojects/arya-helpers'

setup(
    name='aryahelpers',
    packages=find_packages(where='src'),
    package_dir={'': 'src'},
    version='0.1.0',
    description='Arya Helpers Modules',
    long_description='Helper Python modules for Arya + general projects',
    author='Ratnadip Adhikari',
    author_email='ratnadip.adhikari@leoforce.com',
    url=BASE_REPO + '.git',
    download_url=BASE_REPO,
    keywords=['utilities', 'arya', 'helper modules'],
    classifiers=[
        "Development Status :: 4 - Beta",
        "Intended Audience :: Developers",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3",
        "Topic :: Utilities"
    ],
    platforms=['any'],
    install_requires=[],
    # package_data={'aryahelpers': ['configuration/config.helpers.json']}
    include_package_data=True
)



"""text preprocessing class"""
# -*- coding: utf-8 -*-
# @author Ratnadip Adhikari

import re
import unicodedata
from collections import Counter
from aryahelpers.utils.genericutils import elapsedTime


class PreprocessText():
    """Text preprocessing class"""
    def __init__(self):
        pass  # initialized class with empty constructor
        
    def spaces_clean(self, text_str):
        """Clean tabs/linebreaks/extra spaces in a string"""
        return re.sub(r'\s+', ' ', text_str).strip()

    def ascii_filter(self, text_str):
        """The ascii encoder function for a string"""    
        text_str = unicodedata.normalize('NFKD', text_str).encode('ascii', 'ignore').decode('utf-8', 'ignore')
        return text_str

    def spchars_remove(self, text_str, remove_digits=True):
        """The special characters cleaner function for a string"""
        pattern = r'[^\\\/a-zA-Z\s]+' if remove_digits else r'[^\\\/a-zA-Z0-9\s]+' 
        return re.sub(pattern, '', text_str)

    def stopwords_clean(self, text_str: str, stop_words: list):
        """The stopwords removal function"""
        stopw_dict, words_list = Counter(stop_words), text_str.split()
        text_str = " ".join([wd for wd in words_list if self.spchars_remove(wd.lower()).strip() not in stopw_dict])
        return text_str
    
    def emptyvals_drop(self, text_list: list):
        """Removes null/empty values from a list of text strs"""
        return [s for s in text_list if s and str(s) != 'nan']

    def duplicates_drop(self, text_list: list):
        """Removes duplicates from a list of text strs"""
        return list(dict.fromkeys(text_list))

    @elapsedTime
    def preprocess_text(self, text_str, stop_words=[], remove_spchars=True, remove_digits=True, print_el_time=True):
        """This function preprocesses a text string"""    
        text_str = self.ascii_filter(text_str)
        text_str = self.spchars_remove(text_str, remove_digits) if remove_spchars else text_str
        text_str = self.stopwords_clean(text_str, stop_words)
        text_str = self.spaces_clean(text_str)
        return text_str
    
    @elapsedTime
    def preprocess_textlist(self, text_list, stop_words=[], remove_spchars=True, remove_digits=True,
                            remove_nulls=True, remove_dupli=True, print_el_time=True):
        """This function preprocesses a list of text strs"""
        clean_list = text_list.copy()
        clean_list = [self.ascii_filter(s) for s in clean_list]
        clean_list = [self.spchars_remove(s, remove_digits) for s in clean_list] if remove_spchars else clean_list
        clean_list = [self.stopwords_clean(s, stop_words) for s in clean_list]
        clean_list = [self.spaces_clean(s) for s in clean_list]
        clean_list = self.emptyvals_drop(clean_list) if remove_nulls else clean_list
        clean_list = self.duplicates_drop(clean_list) if remove_dupli else clean_list
        return clean_list
        
